/* ------------------------- pack.c -----------------------------------	*/
/*									*/
/* Author:	Eyal Lebedinsky						*/
/* Date:	25 Mar 1994						*/
/* Version:	1.4.00 (portable)					*/
/*									*/
/* This program is released into the public domain. Use it at		*/
/* your own risk.							*/
/*									*/
/* pack [options] File .... >Script					*/
/*									*/
/* Will generate a batch file for copying the Files to multiple volumes	*/
/* (floppies) of size DeviceSize clusters of size ClusterSize bytes.	*/
/* The first floppy will be numbered FirstDisk, and drive A: or B: will	*/
/* be used. Each volume will be labeled with DiskLabel.			*/
/*									*/
/* Notes:								*/
/*	msdos  uses: echo, rem, pause, del, label, copy			*/
/*	mtools uses: echo, mformat, mlabel, mwrite, read, sync		*/
/*	dir    uses: echo, mkdir, rm, cp, read				*/
/*	raw    uses: none - it creates a text file.			*/
/*									*/
/* Options:								*/
/*	-c	cluster size in bytes					*/
/*			default: 1024					*/
/*	-d	device size (data space) in clusters			*/
/*			default: 1024					*/
/*			special: 360  for  360K 5-1/4"			*/
/*				 720  for  720K 3-1/2"			*/
/*				 1.2  for  1.2M 5-1/4"			*/
/*				 1.44 for 1.44M 3-1/2"			*/
/*	-e	do not pack empty files					*/
/*			default: copy empty files			*/
/*	-f	use drive A:/B:						*/
/*			default: A:					*/
/*	-l	disk label (printf format)				*/
/*			default: DISK%04u				*/
/*			-l will disable labeling			*/
/*	-m	method: raw, msdos, mtools, dir				*/
/*			default: METHOD, or else 'raw'			*/
/*	-n	what number is the first disk				*/
/*			default: 1					*/
/*	-p	pause even in 'dir' mode				*/
/*			default: no pause in 'dir' mode			*/
/*	-q	no 'bell' in batch file					*/
/*			default: beep for disk change			*/
/*	-s	'split' large files					*/
/*			default: skip large files			*/
/*	File...	Files to pack, with wildcards				*/
/*									*/
/* To compile (e.g. MSC8.00, with wildcard expansion):			*/
/*	cl -AL -O2 -Gs -DLABEL=\"LABEL%03u\" pack.c			*/
/*		q:lib\setargv.obj -link /NOE				*/
/*									*/
/* usage: 								*/
/*	>pack *.* > p.bat    (will generate the batch file)		*/
/*	>p		     (will execute the batch file)		*/
/*									*/
/* To do:								*/
/*									*/
/* History:								*/
/*	June  1989	version 1.0					*/
/*		initial release						*/
/*	March 1990	version 1.1					*/
/*		add -n option						*/
/*	Oct 1990	version 1.2					*/
/*		add -l option						*/
/*		minor changes						*/
/*	Jan 1991	version 1.3					*/
/*		add -q option						*/
/*	24 Feb 1991	version 1.3.01					*/
/*		Do not fail if 'all files rejected'			*/
/*		Add short names for standard floppy types		*/
/*	25 Mar 1994	version 1.4.00					*/
/*		Add mtools, raw and dir modes				*/
/*		Default device is now 1.44MB floppy			*/
/*		Add -m option						*/
/*		Add -p option						*/
/*		Add -e option						*/
/*		Extend -f option					*/
/*									*/
/* --------------------------------------------------------------------	*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#define	CLSIZE	1024	/* must be power of 2 */
#define	DSIZE	1423	/* device size in clusters */
#define	ENTRY	struct entry

typedef unsigned char uchar;

#ifndef	LABEL
#define	LABEL	"DISK%04u"
#endif

#define METHOD_RAW	0
#define METHOD_MSDOS	1
#define METHOD_MTOOLS	2
#define METHOD_DIR	3

#ifndef METHOD
#define METHOD		"raw"
#endif

static uchar	*label = LABEL;
static int	quiet = 0;
static int	method = METHOD_RAW;
static int	pause = 0;
static int	do_empty = 1;

ENTRY {
	long	size;
	uchar	*name;
	int	diskno;
};

static void
usage ()
{
	fprintf (stderr, "pack v1.4 Mar/1994 by Eyal Lebedinsky\n");
	fprintf (stderr, "%s%s\n",
		"usage: pack [-cnnn -dn -e -fA:/B: -lDiskLabel -mMethod -niii",
		" -p -q -s] file ... >Script");
	fprintf(stderr, "\t-cnnn   Cluster size nnn bytes (default 1024)\n");
	fprintf(stderr, "\t-dnnn   Device size nnn clusters (default 354)\n");
	fprintf(stderr, "\t        special values: 360, 720, 1.2 , 1.44\n");
	fprintf(stderr, "\t-e      do not pack empty files\n");
	fprintf(stderr, "\t-fTo    pack to A:/B:\n");
	fprintf(stderr, "\t-lsss   use template sss for label (default \"%s\")\n",
		LABEL);
	fprintf(stderr, "\t-l      do not label disks\n");
	fprintf(stderr, "\t-mMethod output method: raw,msdos,mtools,dir (delault %s)\n",
		METHOD);
	fprintf(stderr, "\t-niii   first floppy is number iii (default 1)\n");
	fprintf(stderr, "\t-p      pause in 'dir' mode\n");
	fprintf(stderr, "\t-q      quiet (no bell)\n");
	fprintf(stderr, "\t-s      split large files\n");
	fprintf(stderr, "\tfile... files to pack\n");
	exit (4);
}

static int
cmp (a, b)
ENTRY *a, *b;
{
	if (a->size > b->size)	/* note: decreasing order */
		return -1;
	if (a->size < b->size)
		return 1;
	return 0;
}

static int	/* 0: success	1: failure */
store (files, disks, ndisks)
ENTRY *files;
long	disks[];
int	ndisks;
{
	long	bestsize;
	int	best, i;
	
	/* fit in smallest hole. */

	best = -1;
	
	for (i = 0; i < ndisks; ++i) {
		if (disks[i] >= files->size) {
			if (best < 0 || disks[i] < bestsize) {
				bestsize = disks[i];
				best = i;
			}
		}
	}
	
	if (best < 0)
		return 1;
		
	disks[best] -= files->size;
	files->diskno = best;
	return 0;
}

static void
pack_files (files, nfiles, ndisks, disks, drive, first_disk)
ENTRY	*files;
int	nfiles;
int	ndisks;
long	disks[];
char	*drive;
int	first_disk;
{
	int	i, j, k, n;
	char	dir[256];

    switch (method) {

    case METHOD_RAW:
	printf ("set\t%s\t%u\t%u\n", drive, ndisks, first_disk);
	for (i = 0; i < ndisks; ++i) {
		k = first_disk+i;
		printf ("disk\t%s\t%u\t%u\t%u\t%lu\n",
			drive, i+1, ndisks, k, disks[i]);
		if (label[0] != '\0') {
			printf ("label\t%s\t%u\t%u\t", drive, i+1, k);
			printf (label, k);
			printf ("\n");
		}
		for (n = j = 0; j < nfiles; ++j) {
			if (files[j].diskno == i)
				printf ("file\t%s\t%u\t%u\t%u\t%u\t%s\n",
					drive, i+1, k, j+1, ++n, files[j].name);
		}
	}
	printf ("end\n");
	break;

    case METHOD_MSDOS:
	for (i = 0; i < ndisks; ++i) {
		printf ("rem Disk %u/%u: (%lu bytes free)\n", i+1, ndisks, disks[i]);
		printf ("pause insert disk number %u in drive %s%c\n",
			first_disk + i, drive, quiet ? ' ' : '\a');
		printf ("echo y|del %s\\ >nul\n", drive);
		if (label[0] != '\0') {
			printf ("label %s", drive);
			printf (label, first_disk + i);
			printf ("\n");
		}
		for (j = 0; j < nfiles; ++j) {
			if (files[j].diskno == i)
				printf ("\tcopy %s %s\n", files[j].name, drive);
		}
	}
	break;

    case METHOD_MTOOLS:
	for (i = 0; i < ndisks; ++i) {
		printf ("echo \"Disk %u/%u: (%lu bytes free)\"\n",
			i+1, ndisks, disks[i]);
		printf ("\techo insert disk number %u in drive %s%c;read xyz\n",
			first_disk + i, drive, quiet ? ' ' : '\a');
		printf ("\tmformat %s\n", drive);
		if (label[0] != '\0') {
			printf ("\techo ");
			printf (label, first_disk + i);
			printf ("|mlabel %s >/dev/null\n", drive);
		}
		for (j = 0; j < nfiles; ++j) {
			if (files[j].diskno == i)
				printf ("\tmwrite %s %s\n",
					files[j].name, drive);
		}
		printf ("\tsync\n");
	}
	break;

    case METHOD_DIR:
	for (i = 0; i < ndisks; ++i) {
		k = first_disk+i;
		sprintf (dir, label, k);
		printf ("echo \"Disk %u/%u: (%lu bytes free)\"\n",
			i+1, ndisks, disks[i]);
		if (pause) {
			printf ("\techo insert disk number %u in drive %s%c\n",
				k, drive, quiet ? ' ' : '\a');
			printf ("\tread xyz\n");
		}
		printf ("\tmkdir %s%s\n", drive, dir);
		printf ("\trm %s%s/*\n", drive, dir);
		for (j = 0; j < nfiles; ++j) {
			if (files[j].diskno == i)
				printf ("\tcp %s %s%s\n",
					files[j].name, drive, dir);
		}
	}
	break;
    }
}

static void
split_files (files, drive, first)
ENTRY	*files;
char	*drive;
int	first;
{
	int	i;

    switch (method) {

    case METHOD_RAW:
	for (i = 0; i < first; ++i)
		printf ("split\t%s\t%s\n", drive, files[i].name);
	break;

    case METHOD_MSDOS:
	for (i = 0; i < first; ++i)
		printf ("split %s%s\n", files[i].name, drive[0] ==  'B' ? " -b" : "");
	break;

    case METHOD_MTOOLS:
	for (i = 0; i < first; ++i)
		printf ("echo split %s %s\n", files[i].name, drive);
	break;
    }
}

static long	/* <0: failure	>=0: success, free space on last disk */
try (files, nfiles, ndisks, drive, device_size, first_disk)
ENTRY	*files;
int	nfiles, ndisks;
char	*drive;
long	device_size;
int	first_disk;
{
	long	*disks, l;
	int	i;

/* allocate array for disks free space list
*/
	if ((disks = (long *) malloc (ndisks * sizeof (*disks))) == NULL) {
		fprintf (stderr, "Not enough memory.\n");
		exit (4);
	}

/* initially the disks are all empty
*/
	for (i = 0; i < ndisks; disks[i++] = device_size);

/* now try storing each file in turn on some disk
*/
	for (i = 0; i < nfiles; ++i) {
		if (store (&files[i], disks, ndisks)) {
			/* release storage and return failure */
			free (disks);
			return -1L;
		}
	}

/* if we got this far then we have a good packing order
*/
	pack_files (files, nfiles, ndisks, disks, drive, first_disk);

/* release storage and return success
*/
	l = disks[ndisks-1];
	free (disks);
	return l;
}

static int
parse_method (s)
char	*s;
{
	if (!strcmp (s, "raw"))
		method = METHOD_RAW;
	else if (!strcmp (s, "msdos"))
		method = METHOD_MSDOS;
	else if (!strcmp (s, "mtools"))
		method = METHOD_MTOOLS;
	else if (!strcmp (s, "dir"))
		method = METHOD_DIR;
	else {
		fprintf (stderr, "invalid method \"%s\"\n", s);
		return (1);
	}
	return (0);
}

void
main (argc, argv)
int   argc;
char  *argv[];
{
	int	i,
		nfiles = 0,
		first,
		last,
		errors = 0,
		first_disk = 1,
		split = 0;
	char	*drive = "A:";
	long	l,
		cluster_size = CLSIZE,
		device_size = DSIZE,
		total = 0L,
		totbig = 0L;
	ENTRY	*files;
	struct stat stat_data;

	if (argc < 2)
		usage ();

	errors += parse_method (METHOD);

	/* allocate maximum size for array of files data */
	if ((files = (ENTRY *) malloc ((argc - 1) * sizeof (*files))) == NULL) {
		fprintf (stderr, "Not enough memory.\n");
		exit (4);
	}

	/* scan arguments for options and files */
	for (i = 1; i < argc; ++i) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'c':
				if (sscanf (&argv[i][2], "%ld", &l) != 1 ||
					l <= 0) {
					fprintf (stderr, "invalid cluster size %s\n",
						argv[i]);
					++errors;
					break;
				}
				cluster_size = l;
				break;

			case 'd':
				if (!strcmp (argv[i]+2, "360")) {
					device_size = 354;
					cluster_size = 1024;
				} else if (!strcmp (argv[i]+2, "720")) {
					device_size = 713;
					cluster_size = 1024;
				} else if (!strcmp (argv[i]+2, "1.2")) {
					device_size = 2371;
					cluster_size = 512;
				} else if (!strcmp (argv[i]+2, "1.44")) {
					device_size = 1423;
					cluster_size = 1024;
				} else if (sscanf (&argv[i][2], "%ld", &l) != 1 ||
					l <= 0) {
					fprintf (stderr, "invalid device size %s\n",
						argv[i]);
					++errors;
					break;
				} else
					device_size = l;
				break;

			case 'e':
				do_empty = 0;
				break;

			case 'f':
				drive = argv[i]+2;
				break;

			case 'l':
				label = argv[i]+2;
				break;

			case 'm':
				errors += parse_method (argv[i]+2);
				break;

			case 'n':
				if (sscanf (&argv[i][2], "%ld", &l) != 1 ||
					l < 0) {
					fprintf (stderr, "invalid first disk %s\n",
						argv[i]);
					++errors;
					break;
				}
				first_disk = l;
				break;

			case 'p':
				pause = 1;
				break;

			case 'q':
				quiet = 1;
				break;

			case 's':
				split = 1;
				break;

			default:
				fprintf (stderr, "invalid option %s\n",
					argv[i]);
				++errors;
				break;
			}
			continue;
		}

		/* get file's stats */
		if ((stat (argv[i], &stat_data)) != 0) {
			fprintf (stderr, "Cannot get size of %s", argv[i]);
			fprintf (stderr, ", skipping file.\n");
			continue;
		}
		if ((stat_data.st_mode & S_IFREG) &&
			(do_empty || stat_data.st_size)) {
			files[nfiles].size = stat_data.st_size;
			files[nfiles].name = argv[i];
			++nfiles;
		}
	}

/* Do some checking.
*/
	switch (method) {

	case METHOD_MSDOS:
	case METHOD_MTOOLS:
		i = drive[0];
		if (!('a' == i || 'A' == i || 'b' == i || 'B' == i) ||
		    !('\0' == drive[1] || (':' == drive[1] && '\0' == drive[2]))) {
			fprintf (stderr, "invalid drive %s\n", drive);
			++errors;
		}
		if ('\0' == drive[1]) {
			if ('a' == i || 'A' == i)
				drive = "A:";
			else
				drive = "B:";
		}
		break;

	case METHOD_DIR:
		if ('\0' == label[0]){ 
			fprintf (stderr, "must have label\n");
			++errors;
		}
		break;
	}
	if (errors)
		usage ();

	device_size *= cluster_size;

	qsort ((void *) (files), (size_t) nfiles, sizeof (*files), cmp);

	first = 0;
	last = first + nfiles;
	for (i = 0; i < nfiles; ++i) {
		files[i].size = ((files[i].size + cluster_size - 1)
				/ cluster_size) * cluster_size;
		if (files[i].size > device_size) {
			fprintf (stderr, "File %s too big (%lu).\n",
				files[i].name, files[i].size);
			totbig += files[i].size;
			++first;
		} else
			total += files[i].size;
	}
	if (first >= last) {
		fprintf (stderr, "All files rejected.\n");
	} else {
		nfiles = last - first;
		for (i = (total - 1) / device_size + 1; i <= nfiles; ++i) {
			l = try (files + first, nfiles, i, drive, device_size,
				first_disk);
			if (l >= 0l)
				break;
		}
		fprintf (stderr, "%u volumes needed (%lu free on last).\n",
				i, l);
	}

	if (split && first > 0) {
		split_files (files, drive, first);
		i = totbig / device_size;
		fprintf (stderr, "%u.%lu volumes needed for large files.\n",
				i, totbig % device_size);
	}

	free (files);	
	exit (0);
}
